import logging
from collections import defaultdict
from ecs_container_exporter.utils import create_metric, TASK_CONTAINER_NAME_TAG, create_task_metrics

log = logging.getLogger()
# Default value is maxed out for some cgroup metrics
CGROUP_NO_VALUE = 0x7FFFFFFFFFFFF000


def calculate_memory_metrics(stats, task_container_tags):
    metrics_by_container = {}
    # task level metrics
    task_metrics = defaultdict(int)
    metric_type = 'gauge'
    for container_id, container_stats in stats.items():
        metrics = []
        tags = task_container_tags[container_id]
        memory_stats = container_stats.get('memory_stats', {})
        log.debug(f'Memory Stats: {container_id} - {memory_stats}')

        for mkey in ['cache']:
            value = memory_stats.get('stats', {}).get(mkey)
            # some values have default garbage value
            if value < CGROUP_NO_VALUE:
                metric = create_metric('mem_' + mkey, value, tags, metric_type)
                metrics.append(metric)
                if mkey == 'cache':
                    task_metrics['mem_cache'] += value

        for mkey in ['usage']:
            value = memory_stats.get(mkey)
            # some values have default garbage value
            if value < CGROUP_NO_VALUE:
                metric = create_metric('mem_' + mkey, value, tags, metric_type)
                metrics.append(metric)
                if mkey == 'usage':
                    task_metrics['mem_usage'] += value

        metrics_by_container[container_id] = metrics

    # task level metrics
    metrics_by_container[TASK_CONTAINER_NAME_TAG] = create_task_metrics(task_metrics, metric_type)

    return metrics_by_container



import requests
import logging
from collections import namedtuple
from datadog.dogstatsd.base import DogStatsd
from prometheus_client.core import CounterMetricFamily, GaugeMetricFamily

# special container name tag for Task level (instead of container) metrics
TASK_CONTAINER_NAME_TAG = '_task_'
METRIC_PREFIX = 'ecs_task'

logging.basicConfig(
    format='%(asctime)s:%(levelname)s:%(message)s',
)

Metric = namedtuple('Metric', ['name', 'value', 'tags', 'type', 'desc'])

statsd = None


def init_statsd_client(statsd_host='localhost', statsd_port=8125):
    global statsd
    if not statsd:
        statsd = DogStatsd(
            statsd_host, statsd_port,
            use_ms=True,
            namespace=METRIC_PREFIX
        )


def get_logger(log_level):
    return logging.getLogger()


def create_metric(name, value, tags, type, desc=''):
    return Metric(name, value, tags, type, desc)


def create_prometheus_metric(metric):
    metric_name = METRIC_PREFIX +'_'+ metric.name
    if metric.type == 'counter':
        pm = CounterMetricFamily(metric_name, metric.desc, labels=metric.tags.keys())
    elif metric.type == 'gauge':
        pm = GaugeMetricFamily(metric_name, metric.desc, labels=metric.tags.keys())
    else:
        raise Exception(f'Unknown metric type: {metric.type}')

    pm.add_metric(labels=metric.tags.values(), value=metric.value)
    return pm


def format_dogstatsd_tags(tags):
    """
    {k: v} => ['k:v']

    """
    return [f'{k}:{v}' for k, v in tags.items()]


def send_statsd(metric):
    global statsd
    tags = format_dogstatsd_tags(metric.tags)
    if metric.type == 'counter':
        statsd.increment(metric.name, metric.value, tags=tags)
    elif metric.type == 'gauge':
        statsd.gauge(metric.name, metric.value, tags=tags)
    else:
        raise Exception(f'Unknown metric type: {metric.type}')


def create_task_metrics(task_metrics, metric_type):
    """
    Convert task_metrics key, value hash to actual metrics

    """
    metrics = []
    tags = task_metric_tags()
    for mkey, mvalue in task_metrics.items():
        metrics.append(
            create_metric(mkey, mvalue, tags, metric_type)
        )

    return metrics


def ecs_task_metdata(url, timeout):
    response = requests.get(url, timeout=timeout)

    if response.status_code != 200:
        raise Exception(f'Error: Non 200 response from url {url}')

    try:
        return response.json()

    except ValueError as e:
        raise Exception(f'Error: decoding json response from url {url} response {response.text}: {e}')


def task_metric_tags():
    """
    Special tag that will apply to `task` level metrics, as opposed to
    `container` level metrics.

    """
    return {'container_name': TASK_CONTAINER_NAME_TAG}



import logging
from collections import defaultdict
from .utils import create_metric, create_task_metrics, TASK_CONTAINER_NAME_TAG

log = logging.getLogger()


def calculate_network_metrics(stats, task_container_tags):
    """
    "networks": {
        "eth1": {
            "rx_bytes": 564655295,
            "rx_packets": 384960,
            "rx_errors": 0,
            "rx_dropped": 0,
            "tx_bytes": 3043269,
            "tx_packets": 54355,
            "tx_errors": 0,
            "tx_dropped": 0
        }
    }

    """
    metrics_by_container = {}
    # task level metrics
    task_metrics = defaultdict(int)
    # assume these will always be gauge
    metric_type = 'gauge'
    for container_id, container_stats in stats.items():
        metrics = []
        tags = task_container_tags[container_id]
        network_stats = container_stats.get('networks')
        if not network_stats:
            continue

        for iface, iface_stats in network_stats.items():
            iface_tag = tags.copy()
            iface_tag['iface'] = iface

            for stat, value in iface_stats.items():
                metrics.append(
                    create_metric('network_' + stat + '_total', value, tags,
                                  metric_type, 'Network '+stat)
                )
                # assume single iface
                task_metrics['network_' + stat + '_total'] += value

        metrics_by_container[container_id] = metrics

    # Task metrics
    metrics_by_container[TASK_CONTAINER_NAME_TAG] = create_task_metrics(task_metrics, metric_type)

    return metrics_by_container



import logging
from collections import defaultdict
from ecs_container_exporter.utils import create_metric, create_task_metrics, TASK_CONTAINER_NAME_TAG

log = logging.getLogger()


def calculate_io_metrics(stats, task_container_tags):
    """
    Calculate IO metrics from the below data:

    "blkio_stats": {
        "io_merged_recursive": [],
        "io_queue_recursive": [],
        "io_service_bytes_recursive": [
            {
                "major": 259,
                "minor": 0,
                "op": "Read",
                "value": 10653696
            },
            {
                "major": 259,
                "minor": 0,
                "op": "Write",
                "value": 0
            },
            {
                "major": 259,
                "minor": 0,
                "op": "Sync",
                "value": 10653696
            },
            {
                "major": 259,
                "minor": 0,
                "op": "Async",
                "value": 0
            },
            {
                "major": 259,
                "minor": 0,
                "op": "Total",
                "value": 10653696
            }
        ],
        "io_service_time_recursive": [],
        "io_serviced_recursive": [
            {
                "major": 259,
                "minor": 0,
                "op": "Read",
                "value": 164
            },
            {
                "major": 259,
                "minor": 0,
                "op": "Write",
                "value": 0
            },
            {
                "major": 259,
                "minor": 0,
                "op": "Sync",
                "value": 164
            },
            {
                "major": 259,
                "minor": 0,
                "op": "Async",
                "value": 0
            },
            {
                "major": 259,
                "minor": 0,
                "op": "Total",
                "value": 164
            }
        ],
        "io_time_recursive": [],
        "io_wait_time_recursive": [],
        "sectors_recursive": []
    },
    """
    metrics_by_container = {}
    # task level metrics
    task_metrics = defaultdict(int)
    # assume these will always be gauge
    metric_type = 'gauge'
    for container_id, container_stats in stats.items():
        metrics = []
        blkio_stats = container_stats.get('blkio_stats')

        iostats = {'io_service_bytes_recursive': 'bytes', 'io_serviced_recursive': 'iops'}
        for blk_key, blk_type in iostats.items():
            tags = task_container_tags[container_id]
            read_counter = write_counter = 0
            for blk_stat in blkio_stats.get(blk_key):
                if blk_stat['op'] == 'Read' and 'value' in blk_stat:
                    read_counter += blk_stat['value']
                elif blk_stat['op'] == 'Write' and 'value' in blk_stat:
                    write_counter += blk_stat['value']

            metrics.append(
                create_metric('disk_read_' + blk_type + '_total', read_counter, tags,
                              metric_type, 'Total disk read ' + blk_type)
            )
            metrics.append(
                create_metric('disk_written_' + blk_type + '_total', write_counter, tags,
                              metric_type, 'Total disk written ' + blk_type)
            )
            task_metrics['disk_read_' + blk_type + '_total'] += read_counter
            task_metrics['disk_written_' + blk_type + '_total'] += write_counter

        metrics_by_container[container_id] = metrics

    # task level metrics
    metrics_by_container[TASK_CONTAINER_NAME_TAG] = create_task_metrics(task_metrics, metric_type)

    return metrics_by_container



import logging
from ecs_container_exporter.utils import create_metric, task_metric_tags, TASK_CONTAINER_NAME_TAG

log = logging.getLogger(__name__)


def calculate_cpu_metrics(docker_stats, task_cpu_limit, task_container_limits,
                          task_container_tags):
    """
    Calculate cpu metrics based on the docker container stats json.

    The precpu_stats is calculated by docker stats api at 1s interval,
    which is too small, instead we scample twice over a longer interval.
    https://github.com/moby/moby/blob/b50ba3da1239a56456634f74660f43a27df6b3f2/daemon/daemon.go#L1055

    `docker_stats` is a list with two samples over an interval. CPU usage
    is caculated by comparing the `cpu_stats` of the first and second sample.

    Task and Container Cpu usage is `scaled` against the applicable CPU
    limits. This is more accurate than using the actual values which give
    utilization against the Host CPU.

    Specifically calculate based on the following data.

    "cpu_stats": {
        "cpu_usage": {
            "percpu_usage": [
                826860687,
                830807540,
                823365887,
                844077056
            ],
            "total_usage": 3325111170,
            "usage_in_kernelmode": 1620000000,
            "usage_in_usermode": 1600000000
        },
        "online_cpus": 4,
        "system_cpu_usage": 35595977360000000,
        "throttling_data": {
            "periods": 0,
            "throttled_periods": 0,
            "throttled_time": 0
        }
    },

    """
    # Container_id to metrics mapping
    metrics_by_container = {}
    # Total task cpu usage
    task_cpu_usage_ratio = 0.0

    prev_stats = docker_stats[0]
    stats = docker_stats[1]
    online_cpus = get_online_cpus(stats)
    for container_id, container_stats in stats.items():
        metrics = []
        tags = task_container_tags[container_id]

        cpu_stats = container_stats.get('cpu_stats', {})
        prev_cpu_stats = prev_stats[container_id].get('cpu_stats', {})
        log.debug(f'CPU Stats: {container_id} - curr {cpu_stats} - prev {prev_cpu_stats}')

        # `cpu_stats` sometimes maybe empty when container is just started
        if not prev_cpu_stats:
            continue

        usage_delta = cpu_usage_diff(cpu_stats, prev_cpu_stats)
        # system_cpu_usage is sum of `all` cpu usage on the `host`:
        # https://github.com/moby/moby/blob/54d88a7cd366fd8169b8a96bec5d9f303d57c425/daemon/stats/collector_unix.go#L31
        #
        system_delta = system_cpu_usage_diff(cpu_stats, prev_cpu_stats)
        log.debug(f'usage_delta {usage_delta} system_delta {system_delta}')

        cpu_usage_ratio = calculate_cpu_usage(usage_delta, system_delta)
        log.debug(f'Usage Ratio {cpu_usage_ratio} {online_cpus}')
        # add to total task cpu usage
        task_cpu_usage_ratio += cpu_usage_ratio

        # Scale cpu usage against the Task or Container Limit
        container_cpu_limit = task_container_limits[container_id]['cpu']
        value = normalize_cpu_usage(cpu_usage_ratio, online_cpus,
                                    task_cpu_limit,
                                    container_cpu_limit)

        metric = create_metric('cpu_usage_ratio', round(value, 4), tags, 'gauge',
                               'CPU usage ratio')
        metrics.append(metric)

        # Per cpu metrics
        metrics.extend(
            percpu_metrics(cpu_stats, prev_cpu_stats, system_delta, online_cpus,
                           task_cpu_limit, container_cpu_limit, tags)
        )

        metrics.extend(
            cpu_user_kernelmode_metrics(cpu_stats, prev_cpu_stats, tags)
        )

        # CPU throttling
        # TODO: figure out how this will look at the Task level
        metrics.extend(
            cpu_throttle_metrics(cpu_stats, tags)
        )

        metrics_by_container[container_id] = metrics

    # total task cpu ratio
    if task_cpu_limit != 0:
        task_cpu_usage_ratio = scale_cpu_usage(task_cpu_usage_ratio, online_cpus, task_cpu_limit)

    tags = task_metric_tags()
    metric = create_metric('cpu_usage_ratio', round(task_cpu_usage_ratio, 4), tags, 'gauge',
                           'CPU usage ratio')

    metrics_by_container[TASK_CONTAINER_NAME_TAG] = [metric]

    return metrics_by_container


def cpu_throttle_metrics(cpu_stats, tags):
    metrics = []
    throttling_data = cpu_stats.get('throttling_data')
    for mkey, mvalue in throttling_data.items():
        metric = create_metric('cpu_throttle_' + mkey, mvalue, tags, 'gauge')
        metrics.append(metric)

    return metrics


def cpu_user_kernelmode_metrics(cpu_stats, prev_cpu_stats, tags):
    # user and kernel mode split
    curr_stats = cpu_stats['cpu_usage']
    prev_stats = prev_cpu_stats['cpu_usage']
    kernelmode_delta = curr_prev_diff(curr_stats, prev_stats, 'usage_in_kernelmode')
    kernelmode_metric = create_metric('cpu_kernelmode', kernelmode_delta, tags, 'gauge',
                                      'cpu usage in kernel mode')

    usermode_delta = curr_prev_diff(curr_stats, prev_stats, 'usage_in_usermode')
    usermode_metric = create_metric('cpu_usermode', usermode_delta, tags, 'gauge',
                                    'cpu usage in user mode')

    return (kernelmode_metric, usermode_metric)


def percpu_metrics(cpu_stats, prev_cpu_stats, system_delta, online_cpus,
                   task_cpu_limit, container_cpu_limit, tags):
    metrics = []
    percpu_usage = cpu_stats.get('cpu_usage', {}).get('percpu_usage', [])
    prev_percpu_usage = prev_cpu_stats.get('cpu_usage', {}).get('percpu_usage', [])

    if percpu_usage and prev_percpu_usage:
        for i, value in enumerate(percpu_usage):
            # skip inactive Cpus - https://github.com/torvalds/linux/commit/5ca3726
            if value != 0:
                cpu_tags = tags.copy()
                cpu_tags['cpu'] = str(i)
                if prev_percpu_usage[i] and system_delta:
                    usage_delta = float(value) - float(prev_percpu_usage[i])
                    usage_ratio = usage_delta / system_delta
                else:
                    usage_ratio = 0.0

                value = normalize_cpu_usage(usage_ratio, online_cpus,
                                            task_cpu_limit,
                                            container_cpu_limit)

                metric = create_metric('percpu_usage_ratio', round(usage_ratio, 4), cpu_tags, 'gauge',
                                       'Per CPU usage ratio')
                metrics.append(metric)

    return metrics


def calculate_cpu_usage(usage_delta, system_delta):
    if usage_delta and system_delta:
        # Keep it to 100% instead of scaling by number of cpus :
        # https://github.com/moby/moby/issues/29306#issuecomment-405198198
        #
        cpu_usage_ratio = usage_delta / system_delta
    else:
        cpu_usage_ratio = 0.0

    return cpu_usage_ratio


def system_cpu_usage_diff(cpu_stats, prev_cpu_stats):
    return curr_prev_diff(cpu_stats, prev_cpu_stats, 'system_cpu_usage')


def cpu_usage_diff(cpu_stats, prev_cpu_stats):
    return curr_prev_diff(cpu_stats['cpu_usage'], prev_cpu_stats['cpu_usage'],
                          'total_usage')


def curr_prev_diff(curr_stats, prev_stats, metric_key):
    curr = curr_stats[metric_key]
    prev = prev_stats[metric_key]
    if curr and prev:
        return curr - prev
    else:
        return 0.0


def scale_cpu_usage(cpu_usage_ratio, host_cpu_count, cpu_share_limit):
    """
    Scales cpu ratio originally calculated against total host cpu capacity,
    with the corresponding cpu shares limit (at task or container level)

    host_cpu_count is multiplied by 1024 to get the total available cpu shares

    """
    scaled_cpu_usage_ratio = (cpu_usage_ratio * host_cpu_count * 1024) / cpu_share_limit
    return scaled_cpu_usage_ratio


def normalize_cpu_usage(cpu_usage_ratio, host_cpu_count,
                        task_cpu_limit, container_cpu_limit):
    """
    Task Limit - Container Limit - Cpu metric
        0  -  0  -  no scaling
        0  -  x  -  scale against container limit
        x  -  x  -  scale against container limit
        x  -  0  -  scale against task limit

    """
    log.debug(f'Normalize CPU: {cpu_usage_ratio} - {host_cpu_count} - {task_cpu_limit} - {container_cpu_limit}')
    if task_cpu_limit == 0 and container_cpu_limit == 0:
        # no scaling, task can use all of host CPU
        return cpu_usage_ratio

    if container_cpu_limit != 0:
        # case 2 and 3 above
        # this can go above 100% when task_cpu_limit is 0
        #
        return scale_cpu_usage(cpu_usage_ratio, host_cpu_count, container_cpu_limit)

    if task_cpu_limit != 0:
        # case 4, should not go above 100% as cpu usage
        # is limited by task_cpu_limit
        #
        return scale_cpu_usage(cpu_usage_ratio, host_cpu_count, task_cpu_limit)


def get_online_cpus(stats):
    """
    Assume this number is same for all containers

    """
    for container_id, container_stats in stats.items():
        return container_stats['cpu_stats']['online_cpus']



#!/usr/bin/env python3
import sys
import time
import click
import pickle
import signal
from requests.compat import urljoin

from prometheus_client import start_http_server
from prometheus_client.core import REGISTRY

from ecs_container_exporter import utils
from ecs_container_exporter.utils import create_metric
from ecs_container_exporter.cpu_metrics import calculate_cpu_metrics
from ecs_container_exporter.memory_metrics import calculate_memory_metrics
from ecs_container_exporter.io_metrics import calculate_io_metrics
from ecs_container_exporter.network_metrics import calculate_network_metrics

import logging
log = logging.getLogger(__name__)


class ECSContainerExporter(object):

    include_containers = []
    exclude_containers = []
    # 1 - healthy, 0 - unhealthy
    exporter_status = 1
    # initial task metrics that do not change
    static_task_metrics = []
    # individual container tags
    task_container_tags = {}
    # task limits
    task_cpu_limit = 0
    task_mem_limit = 0
    # individual container limits
    task_container_limits = {}
    # the Task level metrics are included by default
    include_container_ids = [utils.TASK_CONTAINER_NAME_TAG]
    # stats are collected and aggregated across this interval
    interval = 60

    def __init__(self, metadata_url=None, include_containers=None, exclude_containers=None, interval=60, http_timeout=60):

        self.task_metadata_url = urljoin(metadata_url + '/', 'task')
        # For testing
        # self.task_stats_url = urljoin(metadata_url + '/', 'stats')
        self.task_stats_url = urljoin(metadata_url + '/', 'task/stats')

        if exclude_containers:
            self.exclude_containers = exclude_containers
        if include_containers:
            self.include_containers = include_containers

        self.interval = interval
        self.http_timeout = http_timeout

        self.log = logging.getLogger(__name__)
        # cacheing mainly for statsd metric type, since prometheus will run this
        # only once its lifetime
        self.cache_file_path = '/tmp/ecs-container-exporter.cache'
        self.collect_static_metrics()

    def start_prometheus_eporter(self, exporter_port):
        self.log.info(f'Exporter initialized with '
                      f'metadata_url: {self.task_metadata_url}, '
                      f'task_stats_url: {self.task_stats_url}, '
                      f'http_timeout: {self.http_timeout}, '
                      f'include_containers: {self.include_containers}, '
                      f'exclude_containers: {self.exclude_containers}')

        REGISTRY.register(self)
        # Start exporter http service
        start_http_server(int(exporter_port))
        while True:
            time.sleep(10)

    def send_statsd_metrics(self, statsd_host, statsd_port):
        utils.init_statsd_client(statsd_host, statsd_port)
        for metric in self.collect_all_metrics():
            utils.send_statsd(metric)

    def cache_load_task_metadata(self):
        try:
            with open(self.cache_file_path, 'rb') as cf:
                return pickle.load(cf)
        except FileNotFoundError:
            pass

        return None

    def cache_write_task_metadata(self, data):
        with open(self.cache_file_path, 'wb') as cf:
            pickle.dump(data, cf)

    def collect_static_metrics(self):
        metadata = self.cache_load_task_metadata()
        if metadata:
            self.log.debug(f'Using cached task metadata response from {self.cache_file_path}')
        else:
            retries = 24
            while True and retries > 0:
                # wait for the task to be in running state
                time.sleep(5)
                retries -= 1
                try:
                    metadata = utils.ecs_task_metdata(self.task_metadata_url, self.http_timeout)
                    self.cache_write_task_metadata(metadata)
                except Exception as e:
                    self.exporter_status = 0
                    self.log.exception(e)
                    continue

                if metadata.get('KnownStatus') != 'RUNNING':
                    self.log.warning(f'ECS Task not yet in RUNNING state, current status is: {metadata["KnownStatus"]}')
                    continue
                else:
                    break

            self.exporter_status = 1

        self.log.debug(f'Discovered Task metadata: {metadata}')
        self.parse_task_metadata(metadata)

    def parse_task_metadata(self, metadata):
        self.static_task_metrics = []
        self.task_container_tags = {}
        self.task_container_limits = {}

        # task cpu/mem limit
        task_tag = utils.task_metric_tags()
        self.task_cpu_limit, self.task_mem_limit = self.cpu_mem_limit(metadata)

        metric = create_metric('cpu_limit', self.task_cpu_limit, task_tag, 'gauge', 'Task CPU limit')
        self.static_task_metrics.append(metric)

        metric = create_metric('mem_limit', self.task_mem_limit, task_tag, 'gauge', 'Task Memory limit')
        self.static_task_metrics.append(metric)

        # container tags and limits
        for container in metadata['Containers']:
            container_id = container['DockerId']
            container_name = container['Name']

            if self.should_process_container(container_name,
                                             self.include_containers,
                                             self.exclude_containers):
                self.log.info(f'Processing stats for container: {container_name} - {container_id}')
                self.include_container_ids.append(container_id)
            else:
                self.log.info(f'Excluding container: {container_name} - {container_id} as per exclusion')

            self.task_container_tags[container_id] = {'container_name': container_name}

            # container cpu/mem limit
            cpu_value, mem_value = self.cpu_mem_limit(container)
            self.task_container_limits[container_id] = {'cpu': cpu_value, 'mem': mem_value}

            if container_id in self.include_container_ids:
                metric = create_metric('cpu_limit', cpu_value, self.task_container_tags[container_id],
                                       'gauge', 'Limit in percent of the CPU usage')
                self.static_task_metrics.append(metric)

                metric = create_metric('mem_limit', mem_value, self.task_container_tags[container_id],
                                       'gauge', 'Limit in memory usage in MBs')
                self.static_task_metrics.append(metric)

    def should_process_container(self, container_name, include_containers, exclude_containers):
        if container_name in exclude_containers:
            return False
        else:
            if include_containers:
                if container_name in include_containers:
                    return True
                else:
                    return False
            else:
                return True

    def cpu_mem_limit(self, metadata):
        cpu_limit = metadata.get('Limits', {}).get('CPU', 0)

        # Can be CPU shares or CPU units
        # normalize to `CPU shares` since 256 is min share
        scale_factor = 1024 if cpu_limit < 256 else 1

        mem_limit = metadata.get('Limits', {}).get('Memory', 0)

        return (
            cpu_limit * scale_factor,
            mem_limit
        )

    # All metrics are collected here
    def collect_all_metrics(self):
        container_metrics = self.collect_container_metrics()

        # exporter status metric
        metric = create_metric('exporter_status', self.exporter_status, {},
                               'gauge', 'Exporter Status')
        container_metrics.append(metric)

        return self.static_task_metrics + container_metrics

    # prometheus exporter collect function
    def collect(self):
        for metric in self.collect_all_metrics():
            yield utils.create_prometheus_metric(metric)

    def collect_container_metrics(self):
        metrics = []
        docker_stats = []
        # collect two stats samples as per the configured interval
        try:
            docker_stats.append(
                utils.ecs_task_metdata(self.task_stats_url, self.http_timeout)
            )
            # ECS stats API returns immediately, but docker stats waits for 1s adjust this accordingly
            time.sleep(self.interval)

            docker_stats.append(
                utils.ecs_task_metdata(self.task_stats_url, self.http_timeout)
            )
            if not all(docker_stats):
                raise Exception('Some stats are empty, try again')

            self.exporter_status = 1

        except Exception as e:
            self.exporter_status = 0
            self.log.exception(e)
            return metrics

        container_metrics_all = self.parse_container_metadata(docker_stats,
                                                              self.task_cpu_limit,
                                                              self.task_container_limits,
                                                              self.task_container_tags)

        # flatten and filter excluded containers
        filtered_container_metrics = []
        for metrics_by_container in container_metrics_all:
            for container_id, metrics in metrics_by_container.items():
                if container_id in self.include_container_ids:
                    filtered_container_metrics.extend(metrics)

        return filtered_container_metrics

    def parse_container_metadata(self, docker_stats, task_cpu_limit,
                                 task_container_limits, task_container_tags):
        """
        More details on the exposed docker metrics
        https://github.com/moby/moby/blob/c1d090fcc88fa3bc5b804aead91ec60e30207538/api/types/stats.go

        """
        container_metrics_all = []

        # ignore stats for containers in STOPPED state
        for container_id in list(docker_stats[1].keys()):
            if not docker_stats[1][container_id]:
                del(docker_stats[0][container_id])
                del(docker_stats[1][container_id])
                self.log.debug(f'Ignoring null stats for container_id {container_id}')

        try:
            # CPU metrics
            container_metrics_all.append(
                calculate_cpu_metrics(docker_stats,
                                      task_cpu_limit,
                                      task_container_limits,
                                      task_container_tags)
            )

            # Memory metrics
            container_metrics_all.append(
                calculate_memory_metrics(docker_stats[1], task_container_tags)
            )

            # I/O metrics
            container_metrics_all.append(
                calculate_io_metrics(docker_stats[1], task_container_tags)
            )

            # network metrics
            container_metrics_all.append(
                calculate_network_metrics(docker_stats[1], task_container_tags)
            )

        except Exception as e:
            self.log.warning("Could not retrieve metrics for {}: {}".format(task_container_tags, e), exc_info=True)
            self.exporter_status = 1

        return container_metrics_all


def shutdown(sig_number, frame):
    log.info("Recevied signal {}, Shuttting down".format(sig_number))
    sys.exit(0)


@click.command()
@click.option('--metadata-url', envvar='ECS_CONTAINER_METADATA_URI', type=str, default=None,
              help='Override ECS Metadata Url')
@click.option('--exporter-port', envvar='EXPORTER_PORT', type=int, default=9545,
              help='Change exporter listen port')
@click.option('--use-statsd', envvar='USE_STATSD', is_flag=True, type=bool, default=False,
              help='Emit metrics to statsd instead of starting Prometheus exporter')
@click.option('--statsd-host', envvar='STATSD_HOST', type=str, default='localhost',
              help='Override Stasd Host')
@click.option('--statsd-port', envvar='STATSD_PORT', type=int, default=8125,
              help='Override Stasd Port')
@click.option('--include', envvar='INCLUDE', type=str, default=None,
              help='Comma seperated list of container names to include, or use env var INCLUDE')
@click.option('--exclude', envvar='EXCLUDE', type=str, default=None,
              help='Comma seperated list of container names to exclude, or use env var EXCLUDE')
@click.option('--interval', envvar='INTERVAL', type=int, default=60,
              help='Stats collection and aggregation interval in seconds (specifically for CPU stats)')
@click.option('--log-level', envvar='LOG_LEVEL', type=str, default='INFO',
              help='Log level, default: INFO')
def main(
    metadata_url=None, exporter_port=9545, use_statsd=False, statsd_port=8125, statsd_host='localhost',
    include=None, exclude=None, interval=60, log_level='INFO'
):
    if not metadata_url:
        sys.exit('AWS environment variable ECS_CONTAINER_METADATA_URI not found '
                 'nor is --metadata-url set')

    signal.signal(signal.SIGTERM, shutdown)
    signal.signal(signal.SIGINT, shutdown)

    logging.basicConfig(
        format='%(asctime)s:%(levelname)s:%(message)s',
    )
    logging.getLogger().setLevel(
        getattr(logging, log_level.upper())
    )

    if exclude:
        exclude=exclude.strip().split(',')
    if include:
        include=include.strip().split(',')

    collector = ECSContainerExporter(metadata_url=metadata_url,
                                     include_containers=include,
                                     exclude_containers=exclude,
                                     interval=interval)

    if use_statsd:
        collector.send_statsd_metrics(statsd_host, statsd_port)
    else:
        collector.start_prometheus_eporter(exporter_port)


if __name__ == '__main__':
    main()



import re
import ast
from setuptools import setup


_version_re = re.compile(r'__version__\s+=\s+(.*)')

with open('ecs_container_exporter/__init__.py', 'rb') as f:
    version = str(ast.literal_eval(_version_re.search(f.read().decode('utf-8')).group(1)))

with open('./README.md') as f:
    long_desc = f.read()


setup(
    name='ecs-container-exporter',
    version=version,
    author='Raghu Udiyar',
    author_email='raghusiddarth@gmail.com',
    url='https://github.com/Spaced-Out/ecs-container-exporter',
    license='MIT License',
    description='Prometheus exporter for AWS ECS Task and Container Metrics',
    long_description=long_desc,
    long_description_content_type="text/markdown",
    install_requires=['setuptools>=36.0.0',
                      'Click==7.0',
                      'prometheus-client==0.7.1',
                      'requests==2.22.0',
                      'datadog==0.39.0'],
    packages=['ecs_container_exporter'],
    entry_points={
        'console_scripts': [
            'ecs-container-exporter = ecs_container_exporter.main:main'
        ]
    },
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ]
)
